package com.ys.product.ysmq.front.msg;

import java.io.IOException;
import java.util.List;

import com.ys.product.ysmq.front.enums.ConsumerOperationResultCodeEnum;
import com.ys.product.ysmq.front.enums.PreCommitEnum;
import com.ys.product.ysmq.front.exceptions.YsmqConsumerMessageException;
import com.ys.product.ysmq.front.exceptions.YsmqHttpException;
import com.ys.product.ysmq.front.log.Log;
import com.ys.product.ysmq.front.log.LogFactory;
import com.ys.product.ysmq.front.model.BaseResultResponse;
import com.ys.product.ysmq.front.model.Const;
import com.ys.product.ysmq.front.model.ConsumerInfo;
import com.ys.product.ysmq.front.model.GatewayError;
import com.ys.product.ysmq.front.model.ReadMsgResult;
import com.ys.product.ysmq.front.utils.OkHttpHelper;
import com.ys.product.ysmq.front.utils.ResponseParseUtil;

public abstract class AbstractConsumerMessage {

    private static Log log = LogFactory.getLog(AbstractConsumerMessage.class);

    protected String path = "https://open.ys7.com";

    private String openUrl = path + "/api/lapp/token/get";

    protected String createConsumerUrl = path + "/api/lapp/mq/v1/consumer/";

    protected String getMessagesUrl = path + "/api/lapp/mq/v1/consumer/messages";

    protected String commitUrl = path + "/api/lapp/mq/v1/consumer/offsets";

    // 设置自己的appkey
    protected String appKey = "";

    // 设置自己的appSecret
    protected String appSecret = "";

    // 消费间隔时间ms
    protected static long consumeIntervalTime = 1500;

    protected static String accessToken = null;

    protected ConsumerMessageClient client = null;

    private String consumerId = null;

    protected String group = "group1";

    public AbstractConsumerMessage() {

    }

    // String appKey, String appSecret, String group
    public AbstractConsumerMessage(String appKey, String appSecret, String group) {
        this.appKey = appKey;
        this.appSecret = appSecret;
        this.group = group;
    }

    public AbstractConsumerMessage(String path, String appKey, String appSecret, String group) {
        this(appKey, appSecret, group);
        this.path = path;
        openUrl = path + "/api/lapp/token/get";
        createConsumerUrl = path + "/api/lapp/mq/v1/consumer/";
        getMessagesUrl = path + "/api/lapp/mq/v1/consumer/messages";
        commitUrl = path + "/api/lapp/mq/v1/consumer/offsets";

    }

    public String getPath() {
        return path;
    }

    /**
     * <p>
     * 默认的path= https://open.ys7.com/api/lapp/mq/, 如果需要自行修改
     * </p>
     * 
     * @author jiangyuechao 2018年5月30日 下午4:54:42
     * @param path
     */
    public void setPath(String path) {
        this.path = path;
    }

    public String getOpenUrl() {
        return openUrl;
    }

    /**
     * <p>
     * 默认的openUrl,https://open.ys7.com/api/lapp/token/get. 用于获取accessToken,如有需要自行修改,
     * </p>
     * 
     * @param openUrl
     */
    public void setOpenUrl(String openUrl) {
        this.openUrl = openUrl;
    }

    /**
     * <p>
     * 第一步,获取token,同一个userId可以共用一个token
     * </p>
     */
    public synchronized String getAccessToken() {
        // 第一步,获取token,同一个userId可以共用一个token
        try {
            accessToken = OkHttpHelper.getAccessToken(openUrl, appKey, appSecret);
            log.info("accessToken:{}", accessToken);
            return accessToken;
        } catch (Exception e) {// IOException
            log.error("网络异常.", e);
            throw new YsmqHttpException(e);
        }
    }

    /**
     * <p>
     * 用于accessToken过期的情况,当accessToken过期时,设置为null即可获取新的accessToken.
     * </p>
     * 
     * @author jiangyuechao 2018年6月1日 下午7:46:52
     * @param accessToken
     */
    public static void setAccessToken(String accessToken) {
        AbstractConsumerMessage.accessToken = accessToken;
    }

    /**
     * <p>
     * 创建消费者,一个group对应多个consumer,但是一个group中只有一个消费者能获取到消息,建议一个group一个consumer.
     * 关于组的概念,可以参照kafka组的概念,或者看开放平台的文档.
     * </p>
     * 
     * @param group
     *            组名,group1->group5
     * @return consumerId
     * @throws IOException
     */
    protected String createConsumer() throws IOException {
        String consumerId = null;
        // 1.创建消费者,每调用一次返回一个消费者,最多创建10个消费者.
        BaseResultResponse baseResult = client.createConsumer(group);
        if (baseResult.getCode().equals(ConsumerOperationResultCodeEnum.SUCCESS.getCode())) {// 创建成功
            consumerId = ResponseParseUtil.parseBaseResultResponse(baseResult, Const.CONSUMERID);
            // 设置client的consumerId,consumerId相当于跟消息管道的会话id
            ConsumerInfo consumerInfo = new ConsumerInfo(consumerId, group);
            this.consumerId = consumerId;
            client.setConsumerInfo(consumerInfo);
            log.info("consumerId:{}", consumerId);
        } else {// 对业务异常的处理
            createConsumerErrorHandle(baseResult);
        }
        return consumerId;
    }

    /**
     * <p>
     * 只是读消息,没有提交
     * </p>
     * 
     * @throws IOException
     */
    protected ReadMsgResult simpleConsumerMessage() throws IOException {
        try {
            String msg = client.getMessages(PreCommitEnum.CLOSE_PRECOMMIT.getCode());
            // log.debug("simpleConsumerMessage,cosumerInfo:{},msg:{}", client.getConsumerInfo(),
            // msg);
            return handleMessage(msg, PreCommitEnum.CLOSE_PRECOMMIT.getCode());
        } catch (IOException e) {
            log.error("网络异常.", e);
            throw e;
        }
    }

    /**
     * <p>
     * 读提交模式
     * </p>
     * 
     * @throws IOException
     */
    protected ReadMsgResult readCommitMessage() throws IOException {
        try {
            String msg = client.getMessages(PreCommitEnum.OPEN_PRECOMMIT.getCode());
            // log.debug("readCommitMessage,cosumerInfo:{},msg:{}", client.getConsumerInfo(), msg);
            return handleMessage(msg, PreCommitEnum.OPEN_PRECOMMIT.getCode());
        } catch (IOException e) {
            log.error("网络异常.", e);
            throw e;
        }
    }

    /**
     * <p>
     * 把GatewayError类转换为ReadMsgResult
     * </p>
     */
    protected ReadMsgResult gatewayErrorToReadMsgResult(GatewayError gatewayError) {
        if (gatewayError == null) {
            return null;
        }
        ReadMsgResult msgResult = new ReadMsgResult(gatewayError.getCode(), gatewayError.getMsg(), null);
        return msgResult;
    }

    /**
     * <p>
     * 把GatewayError类转换为BaseResultResponse
     * </p>
     * 
     * @param gatewayError
     * @return
     */
    protected BaseResultResponse gatewayErrorToBaseResultResponse(GatewayError gatewayError) {
        if (gatewayError == null) {
            return null;
        }
        BaseResultResponse msgResult = new BaseResultResponse(gatewayError.getCode(), gatewayError.getMsg(), null);
        return msgResult;
    }

    /**
     * <p>
     * 处理读到的消息,把消息转化为ReadMsgResult
     * </p>
     * 
     * @param msg
     * @return
     */
    protected ReadMsgResult handleMessage(String msg, String pattern) {
        ReadMsgResult msgResult = ResponseParseUtil.parseBaseRecordResult(msg);
        if (msgResult.getCode() == null) {// 此时返回的格式不符合ReadMsgResult
            GatewayError gatewayError = ResponseParseUtil.parseGatewayError(msg);
            log.error("gatewayError,cosumerInfo:{},gatewayError:{}", client.getConsumerInfo(), gatewayError);
            return gatewayErrorToReadMsgResult(gatewayError);
        } else {
            // 读提交模式
            if (pattern.equals(PreCommitEnum.OPEN_PRECOMMIT.getCode()) && isCorrectRead(msgResult.getCode())) {
                reSetCommitFailedRetryCount();
            }
        }
        return msgResult;
    }

    /**
     * <p>
     * 消息消费完,提交offsets
     * </p>
     * 
     * @return 1-成功,0-失败
     * @throws IOException
     */
    protected BaseResultResponse commitOffsets() throws IOException {
        try {
            String msg = client.commitOffsets();
            log.debug("commitOffsets,cosumerInfo:{},msg:{}", client.getConsumerInfo(), msg);
            BaseResultResponse baseResult = ResponseParseUtil.parse(msg);
            if (baseResult == null) {// 此时返回的格式不符合BaseResultResponse
                GatewayError gatewayError = ResponseParseUtil.parseGatewayError(msg);
                log.error("gatewayError,cosumerInfo:{},gatewayError:{}", client.getConsumerInfo(), gatewayError);
                if (gatewayError == null) {
                    log.error("commitOffsets error back,cosumerInfo:{},msg:{}", client.getConsumerInfo(), msg);
                    throw new YsmqHttpException(msg);
                }
                return gatewayErrorToBaseResultResponse(gatewayError);
            } else {
                if (isCorrectCommit(baseResult.getCode())) {
                    reSetCommitFailedRetryCount();
                }
            }
            return baseResult;
        } catch (IOException e) {
            log.error("网络异常.", e);
            throw e;
        }
    }

    /**
     * <p>
     * 创建和提交offset的错误码处理
     * </p>
     * 
     * @param baseResult
     * @return
     */
    protected String buildCreateAndCommitErrorMessage(BaseResultResponse baseResult) {
        StringBuilder builder = new StringBuilder();
        builder.append("code:").append(baseResult.getCode()).append(" message:").append(baseResult.getMsg());
        return builder.toString();
    }

    /**
     * <p>
     * 子类实现它,自定义异常错误码处理,这个返回异常看错误码说明.
     * 比如:70117-consumer数量超出上限,要么用已经创建好的consumerId,要么等5分钟cosnumerId过期(这期间consumerId不能访问消息管道,否则会刷新过期时间).
     * 目前创建消费者,用户需要关注的错误码:70117,70110,70109.
     * </p>
     * 
     * @param baseResult
     */
    protected abstract String createConsumerErrorHandle(BaseResultResponse baseResult) throws IOException;

    /**
     * <p>
     * 读取消息出现的错误码处理
     * </p>
     * 
     * @param baseResult
     */
    protected abstract ReadMsgResult readMessageErrorHandle(ReadMsgResult readMsgResult) throws IOException;

    /**
     * <p>
     * 提交消息出现的错误码处理
     * </p>
     * 
     * @param baseResult
     */
    protected abstract int commitMessageErrorHandle(BaseResultResponse baseResult) throws IOException;

    /**
     * <p>
     * 重置提交失败的重试次数
     * </p>
     */
    protected abstract void reSetCommitFailedRetryCount();

    // 设置处理消息的回调

    /**
     * <p>
     * 是否是正确的读取消息: 正常的返回200 和 70120超时回滚都属于正常.
     * </p>
     * 
     * @return
     */
    protected static boolean isCorrectRead(String code) {
        if (code.equals(ConsumerOperationResultCodeEnum.SUCCESS.getCode())
                || code.equals(ConsumerOperationResultCodeEnum.REPEAT_READ.getCode())) {
            return true;
        }
        return false;
    }

    /**
     * <p>
     * 是否是正确的 提交消息: 正常的返回200.
     * </p>
     * 
     * @return
     */
    protected static boolean isCorrectCommit(String code) {
        if (code.equals(ConsumerOperationResultCodeEnum.SUCCESS.getCode())) {
            return true;
        }
        return false;
    }

    public void setAppKey(String appKey) {
        this.appKey = appKey;
    }

    public void setAppSecret(String appSecret) {
        this.appSecret = appSecret;
    }

    public String getGroup() {
        return group;
    }

    public void setGroup(String group) {
        this.group = group;
    }

    public String getConsumerId() {
        return consumerId;
    }

    /**
     * <p>
     * 消费消息的回调
     * </p>
     */
    public interface ConsumerCallBack {
        void consumerCall(List<Object> msg);
    }

    ConsumerCallBack consumerCallBack;

    public void setConsumerCallBack(ConsumerCallBack consumerCallBack) {
        this.consumerCallBack = consumerCallBack;
    }

    public void consumerCall(List<Object> msg) {
        if (consumerCallBack == null) {
            throw new YsmqConsumerMessageException("回调函数不可为空,请设置回调函数!");
        }
        this.consumerCallBack.consumerCall(msg);
    }
}
