package com.ys.product.ysmq.front.msg;

import java.io.IOException;

import com.ys.product.ysmq.front.exceptions.YsmqHttpException;
import com.ys.product.ysmq.front.log.Log;
import com.ys.product.ysmq.front.log.LogFactory;
import com.ys.product.ysmq.front.model.BaseResultResponse;
import com.ys.product.ysmq.front.model.Const;
import com.ys.product.ysmq.front.model.ConsumerInfo;
import com.ys.product.ysmq.front.utils.OkHttpHelper;
import com.ys.product.ysmq.front.utils.ResponseParseUtil;

import okhttp3.FormBody;
import okhttp3.OkHttpClient;
import okhttp3.Request;
import okhttp3.RequestBody;
import okhttp3.Response;

/**
 * <p>
 * 消费消息的客户端;对消费消息做了封装,在内部封装了consumerId,一个ConsumerMessageClient代表一个消费消息的客户端
 * </p>
 * 
 * @author jiangyuechao 2018年1月24日 上午10:59:18
 * @version V1.0
 * @modificationHistory=========================逻辑或功能性重大变更记录
 * @modify by user: {修改人} 2018年1月24日
 * @modify by reason:{方法名}:{原因}
 */
public class ConsumerMessageClient {

    private static Log log = LogFactory.getLog(ConsumerMessageClient.class);

    /**
     * 创建消费者的url
     */
    private String createConsumerUrl;

    /**
     * 消费消息的url
     */
    private String getMessagesUrl;

    /**
     * 提交确认的url
     */
    private String commitUrl;

    private String accessToken;

    /**
     * 保存消费者id,代表与消息管道的连接
     */
    private ConsumerInfo consumerInfo;

    public ConsumerMessageClient() {
    }

    public ConsumerMessageClient(String createConsumerUrl, String getMessagesUrl, String commitUrl,
        String accessToken) {
        this.createConsumerUrl = createConsumerUrl;
        this.getMessagesUrl = getMessagesUrl;
        this.accessToken = accessToken;
        this.commitUrl = commitUrl;
    }

    /**
     * <p>
     * 创建消费者
     * </p>
     * 
     * @param group
     *            消费者对应的分组,目前固定为group1->5
     * @return 服务返回的相应,网络异常返回null
     * @throws IOException
     */
    public BaseResultResponse createConsumer(String group) throws IOException {
        String url = createConsumerUrl + group;
        BaseResultResponse baseResult = null;
        OkHttpClient client = OkHttpHelper.getOkHttpClient();
        RequestBody body = new FormBody.Builder().add(Const.ACCESSTOKEN, accessToken).build();
        Request request = new Request.Builder().url(url).post(body).build();

        try (Response response = client.newCall(request).execute()) {
            if (response.code() == 200) {
                String data = response.body().string();
                baseResult = ResponseParseUtil.parse(data);
            } else {
                // 打印错误
                log.error("http没有正常返回,code:{},body:{}", response.code(), response.body().string());
                throw new YsmqHttpException("code:" + response.code() + ", body:" + response.body().string());
            }
            return baseResult;
        } catch (IOException e) {
            log.error("创建consumerId异常.", e);
            throw e;
        }
    }

    /**
     * <p>
     * 刷新客户端的accessToken
     * </p>
     * 
     * @param accessToken
     */
    public void refreshAccessToken(String accessToken) {
        if (accessToken != null) {
            this.accessToken = accessToken;
            log.info("刷新accessToken:{}", accessToken);
        }
    }

    /**
     * <p>
     * 消费消息,
     * </p>
     * 
     * @author jiangyuechao 2018年2月6日 下午2:18:09
     * @param preCommit
     *            是否开启读取-提交模式的开关,0-不开启,1-开启
     * @return 网络异常时返回null
     * @throws IOException
     */
    public String getMessages(String preCommit) throws IOException {
        String messages = run(getMessagesUrl, consumerInfo.getConsumerId(), preCommit);
        return messages;
    }

    /**
     * <p>
     * 提交offsets
     * </p>
     * 
     * @author jiangyuechao 2018年1月24日 上午11:21:22
     * @return
     * @throws IOException
     */
    public String commitOffsets() throws IOException {
        // 提交时,没有preCommit参数,设置为null.
        String back = run(commitUrl, consumerInfo.getConsumerId(), null);
        return back;
    }

    /**
     * <p>
     * 执行http请求
     * </p>
     * 
     * @author jiangyuechao 2018年2月6日 下午2:25:20
     * @param url
     *            访问的url
     * @param consumerId
     *            消费者id
     * @param preCommit
     *            读取消息时,开启读取提交模式的开关
     * @return 返回http响应
     * @throws IOException
     */
    public String run(String url, String consumerId, String preCommit) throws IOException {
        OkHttpClient client = OkHttpHelper.getOkHttpClient();
        RequestBody body = null;
        if (preCommit != null && !preCommit.equals("")) {
            body = new FormBody.Builder().add(Const.ACCESSTOKEN, accessToken).add(Const.CONSUMERID, consumerId)
                    .add(Const.PRECOMMIT, preCommit).build();
        } else {
            body = new FormBody.Builder().add(Const.ACCESSTOKEN, accessToken).add(Const.CONSUMERID, consumerId).build();
        }
        Request request = new Request.Builder().url(url).post(body).build();
        try (Response response = client.newCall(request).execute()) {
            if (response.code() == 200) {
                String data = response.body().string();
                return data;
            } else {
                // 打印错误
                log.error("http没有正常返回,code:{},body:{}", response.code(), response.body().string());
                throw new YsmqHttpException("response.code=" + response.code());
            }
        }
    }

    public String getCreateConsumerUrl() {
        return createConsumerUrl;
    }

    public void setCreateConsumerUrl(String createConsumerUrl) {
        this.createConsumerUrl = createConsumerUrl;
    }

    public String getGetMessagesUrl() {
        return getMessagesUrl;
    }

    public void setGetMessagesUrl(String getMessagesUrl) {
        this.getMessagesUrl = getMessagesUrl;
    }

    public String getCommitUrl() {
        return commitUrl;
    }

    public void setCommitUrl(String commitUrl) {
        this.commitUrl = commitUrl;
    }

    public ConsumerInfo getConsumerInfo() {
        return consumerInfo;
    }

    public void setConsumerInfo(ConsumerInfo consumerInfo) {
        this.consumerInfo = consumerInfo;
    }

    @Override
    public String toString() {
        return "ConsumerMessageClient [createConsumerUrl=" + createConsumerUrl + ", getMessagesUrl=" + getMessagesUrl
                + ", commitUrl=" + commitUrl + ", accessToken=" + accessToken + ", consumerInfo=" + consumerInfo + "]";
    }

}
