package com.ys.product.ysmq.front.msg;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import com.ys.product.ysmq.front.log.Log;
import com.ys.product.ysmq.front.log.LogFactory;
import com.ys.product.ysmq.front.strategy.HandleMessageErrorStrategy;
import com.ys.product.ysmq.front.utils.NamedThreadFactory;

public class MultiThreadReadCommitMessage extends ReadCommitMessageThread {
    private static Log log = LogFactory.getLog(MultiThreadReadCommitMessage.class);

    private ExecutorService executorService;

    // Executors;
    private NamedThreadFactory namedThreadFactory;

    /**
     * 线程池大小
     */
    private static int defaultPoolSize = 4;

    private int defaultBatchSize = 100;

    /**
     * 默认一批消息处理超时时间,单位:ms.
     */
    private long defaultTimeOut = 5000;

    private HandleMessageErrorStrategy handleMessageErrorStrategy;

    List<Object> failList = new ArrayList<>();

    public MultiThreadReadCommitMessage(AbstractConsumerMessage consumerMessage, long consumeIntervalTime,
        HandleMessageErrorStrategy handleMessageErrorStrategy) {
        this(consumerMessage, consumeIntervalTime, handleMessageErrorStrategy, defaultPoolSize);
    }

    public MultiThreadReadCommitMessage(AbstractConsumerMessage consumerMessage, long consumeIntervalTime,
        HandleMessageErrorStrategy handleMessageErrorStrategy, int poolSize) {
        super(consumerMessage, consumeIntervalTime);
        namedThreadFactory = new NamedThreadFactory("consumer");
        executorService = new ThreadPoolExecutor(poolSize, poolSize, 0L, TimeUnit.MILLISECONDS,
                new LinkedBlockingQueue<Runnable>(), namedThreadFactory, new ThreadPoolExecutor.CallerRunsPolicy());
        this.handleMessageErrorStrategy = handleMessageErrorStrategy;
    }

    /**
     * 重写为多线程版本,多线程处理策略: 1.对消息按照100一批进行分批处理 2.100以下不分批
     */
    @Override
    protected void consumerMessage(List<Object> data) {
        if (data.size() <= defaultBatchSize) {
            super.consumerMessage(data);
        } else {
            List<Future<Boolean>> futureList = new ArrayList<>();
            Map<Integer, List<Object>> groupMessages = groupMessages(data);
            log.debug("收到消息数:{},分组后的消息,组数:{}", data.size(), groupMessages.size());
            Iterator<Entry<Integer, List<Object>>> iterator = groupMessages.entrySet().iterator();
            while (iterator.hasNext()) {
                final Entry<Integer, List<Object>> next = iterator.next();
                Future<Boolean> submit = executorService.submit(new Callable<Boolean>() {

                    @Override
                    public Boolean call() throws Exception {
                        MultiThreadReadCommitMessage.super.consumerMessage(next.getValue());
                        return true;
                    }
                });
                futureList.add(submit);
            }
            failList.clear();
            int taskCount = futureList.size();
            int failCount = 0;
            for (int i = 0; i < futureList.size(); i++) {
                Future<Boolean> future = futureList.get(i);
                try {
                    // messageTaskTimeout 租户单组消息超时时间
                    Boolean result = future.get(defaultTimeOut, TimeUnit.MILLISECONDS);
                    if (!result) {
                        failCount++;
                        failList.addAll(groupMessages.get(i));
                    }
                } catch (Exception e) {
                    if (e instanceof TimeoutException) {
                        log.error(String.format("线程池执行超时异常,消息数量:{}", groupMessages.get(i).size()), e);
                    } else {
                        log.error(String.format("线程池执行异常,消息数量:{}", groupMessages.get(i).size()), e);
                    }
                } finally {
                    future.cancel(true);
                }
            }
            // 消息处理失败策略:重试or忽略
            if (failCount > 0)
                // 默认是重复消费,建议把消息放到本地消息队列进行处理
                handleMessageErrorStrategy.handleMessageError(failList);
        }
    }

    // 对消息进行分组
    private Map<Integer, List<Object>> groupMessages(List<Object> msgs) {
        int groupSize = msgs.size() / defaultBatchSize;
        Map<Integer, List<Object>> map = new HashMap<>();
        int size = 0;
        int count = 0;
        for (Object obj : msgs) {
            if (map.get(size) == null) {
                List<Object> list = new ArrayList<Object>();
                map.put(size, list);
            }
            map.get(size).add(obj);
            if (size + 1 < groupSize) {
                count++;
                if (count >= 100) {
                    size++;
                    count = 0;
                }
            }
        }
        return map;
    }

    public int getDefaultBatchSize() {
        return defaultBatchSize;
    }

    public void setDefaultBatchSize(int defaultBatchSize) {
        this.defaultBatchSize = defaultBatchSize;
    }

    public long getDefaultTimeOut() {
        return defaultTimeOut;
    }

    public void setDefaultTimeOut(long defaultTimeOut) {
        this.defaultTimeOut = defaultTimeOut;
    }

}
