package com.ys.product.ysmq.front.msg;

import java.io.IOException;

import com.ys.product.ysmq.front.enums.ConsumerOperationResultCodeEnum;
import com.ys.product.ysmq.front.exceptions.YsmqCreateConsumerException;
import com.ys.product.ysmq.front.exceptions.YsmqServerOccuredException;
import com.ys.product.ysmq.front.log.Log;
import com.ys.product.ysmq.front.log.LogFactory;
import com.ys.product.ysmq.front.model.BaseResultResponse;
import com.ys.product.ysmq.front.model.ReadMsgResult;
import com.ys.product.ysmq.front.strategy.HandleMessageErrorStrategy;

/**
 * <p>
 * 实现AbstractConsumerMessage可以自定义读取消息的处理和失败的处理. 这个类是实现了创建,读取,提交失败的处理结果.和读取消息的处理,以供参考.
 * </p>
 * 
 * @author jiangyuechao 2018年2月6日 下午2:27:33
 * @version V1.0
 * @modificationHistory=========================逻辑或功能性重大变更记录
 * @modify by user: {修改人} 2018年2月6日
 * @modify by reason:{方法名}:{原因}
 */
public class StandardConsumerMessage extends AbstractConsumerMessage {
    private static Log log = LogFactory.getLog(StandardConsumerMessage.class);

    /**
     * 提交失败之后,重试次数
     */
    private int commitFailedRetryCount = 0;

    /**
     * 拉取的时间间隔,默认200ms
     */
    long consumeIntervalTime = 200;

    private Thread.UncaughtExceptionHandler exceptionHandler;

    private HandleMessageErrorStrategy messgeStartegy;

    private int poolSize = 4;

    /**
     * 消息分批处理的大小,默认100条消息一批.
     */
    private int batchSize = 100;

    /**
     * 一批消息处理的超时时间
     */
    private long batchMsgTimeOut = 5000;

    public StandardConsumerMessage() {
        super();
    }

    public StandardConsumerMessage(String appKey, String appSecret, String group) {
        super(appKey, appSecret, group);
    }

    // String path
    public StandardConsumerMessage(String path, String appKey, String appSecret, String group) {
        super(path, appKey, appSecret, group);
    }

    public long getConsumeIntervalTime() {
        return consumeIntervalTime;
    }

    /**
     * <p>
     * 设置拉取消息的时间间隔,单位ms
     * </p>
     * 
     * @author jiangyuechao 2018年6月4日 下午9:18:21
     * @param consumeIntervalTime
     */
    public StandardConsumerMessage setConsumeIntervalTime(long consumeIntervalTime) {
        this.consumeIntervalTime = consumeIntervalTime;
        return this;
    }

    public StandardConsumerMessage setIsMultiThread(boolean openMulti) {
        this.isMultiThread = openMulti;
        return this;
    }

    public StandardConsumerMessage setMessageErrorStrategy(HandleMessageErrorStrategy messgeStartegy) {
        this.messgeStartegy = messgeStartegy;
        return this;
    }

    public StandardConsumerMessage setPoolSize(int poolSize) {
        this.poolSize = poolSize;
        return this;
    }

    public StandardConsumerMessage setBatchSize(int batchSize) {
        this.batchSize = batchSize;
        return this;
    }

    /**
     * <p>
     * 设置一批消息处理的耗时时间
     * </p>
     * 
     * @param timeOut
     * @return
     */
    public StandardConsumerMessage setBatchMsgTimeOut(long timeOut) {
        this.batchMsgTimeOut = timeOut;
        return this;
    }

    private boolean isMultiThread = false;

    /**
     * <p>
     * 初始化客户端,设置url,accessToken变量.
     * </p>
     */
    public synchronized void initClient() {
        if (appKey == null || appKey.isEmpty() || appSecret == null || appSecret.isEmpty()) {
            // 可以改为自定义异常
            throw new RuntimeException("appKey或appSecret为空");
        }
        getAccessToken();
        if (client == null) {
            client = new ConsumerMessageClient(createConsumerUrl, getMessagesUrl, commitUrl, accessToken);
        }
        if (isMultiThread) {
            if (messgeStartegy == null) {
                throw new RuntimeException("消息处理失败的策略为空,请设置消息处理失败的处置策略!");
            }
            log.info("开启多线程处理消息模式,线程池大小:{}.", poolSize);
            MultiThreadReadCommitMessage multiConsumerMessage = new MultiThreadReadCommitMessage(this,
                    consumeIntervalTime, messgeStartegy, poolSize);
            multiConsumerMessage.setDefaultBatchSize(batchSize);
            multiConsumerMessage.setDefaultTimeOut(batchMsgTimeOut);
            multiConsumerMessage.start();
        } else {
            log.info("开启单线程处理消息模式");
            // 启动线程:
            ReadCommitMessageThread consumerMessageThread = new ReadCommitMessageThread(this, consumeIntervalTime);
            consumerMessageThread.setUncaughtExceptionHandler(exceptionHandler);
            consumerMessageThread.start();
        }
    }

    /**
     * <p>
     * 设置自定义的捕获消费消息线程抛出的异常
     * </p>
     * 
     * @param exceptionHandler
     */
    public void setUncaughtExceptionHandler(Thread.UncaughtExceptionHandler exceptionHandler) {
        this.exceptionHandler = exceptionHandler;
    }

    /**
     * <p>
     * 创建异常处理,这个返回异常看错误码说明.
     * 比如:70117-consumer数量超出上限,要么用已经创建好的consumerId,要么等5分钟cosnumerId过期(这期间consumerId不能访问消息管道,否则会刷新过期时间).
     * </p>
     * 
     * @param baseResult
     */
    @Override
    protected String createConsumerErrorHandle(BaseResultResponse baseResult) {
        ConsumerOperationResultCodeEnum codeEnum = ConsumerOperationResultCodeEnum.getEnumByValue(baseResult.getCode());
        switch (codeEnum) {
            case ACCESS_TOKEN:
                // 刷新accesstoken
                client.refreshAccessToken(getAccessToken());
                return "";
            // consumer数量超出上限!
            case CREATE_CONSUMER_BAYOND_FAILED:
                log.error("最多只能创建10个consumer,请使已经创建的consumerId.");
                throw new YsmqCreateConsumerException(
                        ConsumerOperationResultCodeEnum.CREATE_CONSUMER_BAYOND_FAILED.getCode(),
                        buildCreateAndCommitErrorMessage(baseResult));
            case PARAM_INVALID:
                log.error("参数不合法,请是用合法的参数.");
                throw new YsmqCreateConsumerException(ConsumerOperationResultCodeEnum.PARAM_INVALID.getCode(),
                        buildCreateAndCommitErrorMessage(baseResult));
            case CONSUMER_NOT_SUBSCRIBED_TOPIC:
                log.error("消费者没有订阅任何topic,如果出现这个错误,请检查appkey是否正确,还不能解决,请联系开放平台.");
                throw new YsmqCreateConsumerException(
                        ConsumerOperationResultCodeEnum.CONSUMER_NOT_SUBSCRIBED_TOPIC.getCode(),
                        buildCreateAndCommitErrorMessage(baseResult));
            default: // 其他错误码 -> 消息管道服务内部出现错误,抛出异常.
                ConsumerOperationResultCodeEnum resultCodeEnum = ConsumerOperationResultCodeEnum
                        .getEnumByValue(baseResult.getCode());
                log.error("code:{},message:{}", resultCodeEnum.getCode(), resultCodeEnum.getMessage());
                // 抛出异常
                throw new YsmqServerOccuredException(buildCreateAndCommitErrorMessage(baseResult));
        }
    }

    @Override
    protected ReadMsgResult readMessageErrorHandle(ReadMsgResult baseResult) throws IOException {
        ConsumerOperationResultCodeEnum codeEnum = ConsumerOperationResultCodeEnum.getEnumByValue(baseResult.getCode());
        switch (codeEnum) {
            case ACCESS_TOKEN:
                client.refreshAccessToken(getAccessToken());
                return simpleConsumerMessage();
            case CONSUMER_NOT_EXIST:
                log.error("consumer不存在.");
                // 创建新的consumer
                createConsumer();
                return simpleConsumerMessage();
            case COMMIT_OFFSET_FAIL:
                log.error("提交offset失败,group:{},consumerId:{}.", group, getConsumerId());
                // 提交失败,使用简单的读消息方法,先读消息,如果还失败,创建新的consumer.
                if (commitFailedRetryCount < 3) {
                    commitFailedRetryCount++;
                    // 读取消息,这里不能使用读提交模式
                    return simpleConsumerMessage();
                } else {
                    commitFailedRetryCount = 0;
                    // 创建新的consumer
                    createConsumer();
                    return simpleConsumerMessage();
                }
            default: // 其他错误码 -> 开放平台+消息管道服务内部出现错误,抛出异常
                StringBuilder builder = new StringBuilder();
                log.error("code:{},message:{}", baseResult.getCode(), baseResult.getMsg());
                builder.append("code:").append(baseResult.getCode()).append(" message:").append(baseResult.getMsg());
                // 抛出异常
                throw new YsmqServerOccuredException(builder.toString());
        }
    }

    /**
     * 返回1,不需要处理. 返回0,需要读取消息
     * 
     * @throws IOException
     */
    @Override
    protected int commitMessageErrorHandle(BaseResultResponse baseResult) throws IOException {

        ConsumerOperationResultCodeEnum codeEnum = ConsumerOperationResultCodeEnum.getEnumByValue(baseResult.getCode());
        switch (codeEnum) {
            case ACCESS_TOKEN:
                client.refreshAccessToken(getAccessToken());
                return 1;
            case CONSUMER_NOT_EXIST:
                log.error("consumer不存在.");
                // 创建新的consumer
                createConsumer();
                return 1;
            case COMMIT_OFFSET_FAIL:
                log.error("提交offset失败,group:{},consumerId:{}.", group, getConsumerId());
                // 提交失败重试三次,第四次创建新的consumer.
                if (commitFailedRetryCount < 3) {
                    commitFailedRetryCount++;
                } else {
                    commitFailedRetryCount = 0;
                    // 创建新的consumer
                    createConsumer();
                }
                return 1;
            default: // 其他错误码 -> 开放平台+消息管道服务内部出现错误,抛出异常
                ConsumerOperationResultCodeEnum resultCodeEnum = ConsumerOperationResultCodeEnum
                        .getEnumByValue(baseResult.getCode());
                log.error("group:{},consumerId:{},code:{},message:{}.", group, getConsumerId(),
                        resultCodeEnum.getCode(), resultCodeEnum.getMessage());
                // 抛出异常
                throw new YsmqServerOccuredException(buildCreateAndCommitErrorMessage(baseResult));
        }
    }

    @Override
    protected void reSetCommitFailedRetryCount() {
        if (commitFailedRetryCount != 0)
            commitFailedRetryCount = 0;
    }

}
